# python imports
import datetime
import os

# django imports
from django.conf import settings
from django.core.cache import cache
from django.db import models
from django.utils.translation import ugettext_lazy as _

# resources imports
from resources.config import CSS
from resources.config import RESOURCE_CHOICES
from resources.config import RESOURCES_DIRECTORY

class MergedResource(models.Model):
    """A merged resource out of one or more registered Resources.

    **Attibute:**

    last_updated
        The date the merged resources has been updated.

    type
        The type of resources which are merged.

    group
        The group resources which are merged.

    filename
        The current filename of the merged resource.

    position
        The position of the merged resource.

    """
    last_updated = models.DateTimeField(_("Last update"), blank=True, null=True)
    type = models.PositiveSmallIntegerField(_("Type"), choices=RESOURCE_CHOICES)
    group = models.CharField(_("Group"), blank=True, max_length=255)
    filename = models.CharField(_("Filename"), blank=True, max_length=40)
    position = models.PositiveIntegerField(default=10)

    def get_path(self):
        """Returns the full path to the resource.
        """
        return os.path.join(RESOURCES_DIRECTORY, self.filename)

    def update(self, text, position):
        """Creates a new resources file and sets text and timestamp.
        """
        import resources.utils
        resources.utils.create_directory()
        filename = self._create_filename()
        new = open(os.path.join(settings.MEDIA_ROOT, RESOURCES_DIRECTORY, filename), "w")
        new.write(text)
        new.close()

        try:
            old_filename = os.path.join(settings.MEDIA_ROOT, self.filename)
            os.unlink(old_filename)
        except OSError:
            pass

        self.last_updated = datetime.datetime.now()
        self.filename = filename
        self.position = position
        self.save()

        cache.set("resourcesmanager-%s" % self.type, self)

    def _create_filename(self):
        suffix = "css" if self.type == CSS else "js"
        if self.group:
            return "%s_%s.%s" % (self.group, datetime.datetime.now().strftime("%Y%m%d%H%M%S"), suffix)
        else:
            return "%s.%s" % (datetime.datetime.now().strftime("%Y%m%d%H%M%S"), suffix)

class Resource(models.Model):
    """A registered resource like CSS or Javascript.

    **Attributes**:

    path
        The path to the resource (without MEDIA_URL).

    position
        The position of the resource within the merged file (important if one
        resource is dependend of another).

    type
        The type of the resource: At the time being: Javascript or CSS.

    minify
        If True the resource is minified.

    media
        The media type of the resource. Only used for CSS resources.

    """
    path = models.CharField(_(u"Script"), max_length=255)
    position = models.PositiveIntegerField(default=10)
    type = models.PositiveSmallIntegerField(choices=RESOURCE_CHOICES)
    merge = models.PositiveSmallIntegerField(default=1)
    minify = models.PositiveSmallIntegerField(default=1)
    group = models.CharField(_(u"Group"), blank=True, max_length=255)
    media = models.CharField(blank=True, max_length=30)

    def __unicode__(self):
        return self.path

    class Meta:
        ordering = ("position", )
        
    def save(self, force_insert=False, force_update=False):
        """
        """
        super(Resource, self).save(force_insert, force_update)
        import resources.utils
        resources.utils.reset()
        resources.utils.create_resources()
        
    def get_path(self):
        """Returns the full path to the resource.
        """
        return os.path.join(RESOURCES_DIRECTORY, self.path)



# python imports
import datetime
import os
import re
import shutil

# django imports
from django.db import connection
from django.db.models import Count
from django.conf import settings
from django.core.cache import cache
from django.shortcuts import _get_queryset

# resources imports
from resources.models import Resource
from resources.models import MergedResource
from resources.config import CSS
from resources.config import JS
from resources.config import RESOURCES_DIRECTORY

# cssutils imports
import cssutils
from cssutils import CSSParser

# jsmin imports
from jsmin import jsmin

def create_resources():
    """Creates merged and minified CSS and Javascript of registered resources.
    """
    create_css()
    create_javascript()

def url_replacer(url, resource):
    if url.startswith("/"):
        return url
    else:
        pos = resource.path.rfind("/")
        path = resource.path[0:pos]
        path = settings.MEDIA_URL + path + "/" + url
        path = re.sub("/+", "/", path)
        return path

def create_css():
    """Creates merged and minified CSS of registered CSS resources.
    """
    groups = get_groups(type=CSS)
    cssutils.ser.prefs.useMinified()

    for group in groups:
        cache.delete("resources-css-%s" % group)
        merged_resource, created = MergedResource.objects.get_or_create(type=CSS, group=group)

        max_position = 0
        text = ""
        for resource in Resource.objects.filter(type=CSS, group=group, merge=True):
            if resource.position > max_position:
                max_position = resource.position

            if resource.media:
                text += " @media %s { " % resource.media

            parser = CSSParser()
            sheet = parser.parseFile(settings.MEDIA_ROOT + resource.path)

            cssutils.replaceUrls(sheet, lambda url: url_replacer(url, resource))

            text += sheet.cssText

            if resource.media:
                text += " } "

        merged_resource.update(text, max_position)

def create_javascript():
    """Creates merged and minified javascript of registered Javascript
    resources.
    """
    cache.delete("resources-js")
    groups = get_groups(type=JS)

    for group in groups:
        cache.delete("resources-js-%s" % group)
        merged_resource, created = MergedResource.objects.get_or_create(type=JS, group=group)

        max_position = 0
        text = ""
        for resource in Resource.objects.filter(type=JS, group=group, merge=True):
            if resource.position > max_position:
                max_position = resource.position
            with open(settings.MEDIA_ROOT + resource.path, "r") as f:
                temp = f.read()

            if resource.minify:
                temp = jsmin(temp)

            text += temp + "\n"

        merged_resource.update(text, max_position)

def register_resource(path, type, merge=True, minify=True, group=None, position=None):
    """Registers a resource.

    **Parameters:**

    path
        The relative path (to MEDIA_URL) to the resource. Must be unique.

    type
        The type of the resource. One of resources.config.CSS,
        resources.config.JS

    merge
        If True the resource will be merged with other resources of the same
        group.

    minify
        If True the resource will be minified.

    group
        The group of the resource. Groups can be merged together and seperately
        included in the HTML page.

    position
        The position of the resource within its group. Lower positions come
        first within HTML.
    """
    resource, created = Resource.objects.get_or_create(path=path, type=type)
    resource.merge = merge
    resource.minify = minify
    resource.group = group
    if position:
        resource.position = position
    resource.save()

def unregister_resource(path):
    """Unregisters the resource.

    **Parameters:**

    path
        The unique path to the resource.
    """
    try:
        resource = Resource.objects.get(path=path)
    except Resource.DoesNotExist:
        pass
    else:
        resource.delete()

def get_cached_object(klass, *args, **kwargs):
    cache_key = "%s-%s" % (klass.__name__.lower(), kwargs.values()[0])
    object = cache.get(cache_key)
    if object is not None:
        return object

    queryset = _get_queryset(klass)

    try:
        object = queryset.get(*args, **kwargs)
    except queryset.model.DoesNotExist:
        return None
    else:
        cache.set(cache_key, object)
        return object

def reset():
    """Re-creates the resources directory.
    """
    MergedResource.objects.all().delete()
    shutil.rmtree(os.path.join(settings.MEDIA_ROOT, RESOURCES_DIRECTORY))
    create_directory()

def create_directory():
    """Creates resources directory.
    """
    try:
        os.mkdir(os.path.join(settings.MEDIA_ROOT, RESOURCES_DIRECTORY))
    except OSError:
        pass

# TODO: Use ORM
def get_groups(type):
    """Returns list of names of all groups.
    """
    cursor = connection.cursor()
    cursor.execute("""SELECT "group"
                      FROM resources_resource
                      WHERE type = %s
                      GROUP BY "group" """ % type)

    return [group[0] for group in cursor.fetchall()]


# django imports
from django.conf import settings
from django.http import HttpResponse

# import resources
import resources.utils

def create_css(request):
    """
    """
    resources.utils.create_css()
    return HttpResponse("Created CSS")

def create_javascript(request):
    """
    """
    resources.utils.create_javascript()
    return HttpResponse("Created Javascript")
    
def create_resources(request):
    resources.utils.reset()
    resources.utils.create_resources()
    return HttpResponse("Created Resources")
    


# python imports
import os

# django imports
from django import template
from django.conf import settings
from django.core.cache import cache
from django.template.loader import render_to_string

# resources imports
from resources.utils import get_cached_object
from resources.config import CSS
from resources.config import JS
from resources.config import RESOURCES_DIRECTORY
from resources.models import Resource
from resources.models import MergedResource

register = template.Library()

@register.inclusion_tag('resources/resources.html', takes_context=True)
def css(context, group=None):
    """Returns HTML to embed generated css..
    
    **Parameters**
    
    group
        The group for which the CSS should be linked in.
    """
    resources = _get_resources(context, type=CSS, group=group)

    cache_key = "resources-css-%s" % group
    result = cache.get(cache_key)
    
    if result:
        return { "resources" : result }
    else:
        # Render the HTML here in order to cache it completely
        result = render_to_string("resources/css.html", resources)
        cache.set(cache_key, result)
        return { "resources" : result }

@register.inclusion_tag('resources/resources.html', takes_context=True)
def javascript(context, group=None):
    """Returns HTML to embed generated javascript. 

    **Parameters**
    
    group
        The group for which the javascript should be linked in.
    """
    resources = _get_resources(context, type=JS, group=group)

    cache_key = "resources-js-%s" % group    
    result = cache.get(cache_key)

    if result:
        return { "resources" : result }
    else:
        # Render the HTML here in order to cache it completely
        result = render_to_string("resources/javascript.html", resources)
        cache.set(cache_key, result)
        return { "resources" : result }

def _get_resources(context, type, group=None):
    MEDIA_URL = context.get("MEDIA_URL", "/")

    if getattr(settings, "RESOURCES_DEBUG", False):
        if group:        
            resources = Resource.objects.filter(type=type, group=group)
        else:
            resources = Resource.objects.filter(type=type)
        return {
            "MEDIA_URL" : MEDIA_URL,
            "resources" : resources,
        }
    else:
        if group:
            resources = list(MergedResource.objects.filter(type=type, group=group))
            resources.extend(Resource.objects.filter(type=type, merge=False, group=group))
        else:
            resources = list(MergedResource.objects.filter(type=type))
            resources.extend(Resource.objects.filter(type=type, merge=False))

        resources.sort(lambda a, b: cmp(a.position, b.position))

        return {
            "MEDIA_URL" : MEDIA_URL,
            "resources" : resources,
        }


# resources imports
# from resources.config import CSS, JS
# from resources.utils import register_resource
# from resources.models import MergedResource
# 
# register_resource(path="/lfc/lightbox/css/jquery.lightbox-0.5.css", type=CSS)
# register_resource(path="/lfc/yui/reset-fonts.css", type=CSS)
# register_resource(path="/lfc/blueprint/src/grid.css", type=CSS)
# register_resource(path="/lfc_theme/css/tiny.css", type=CSS)
# register_resource(path="/lfc_theme/css/lfc.css", type=CSS)
# 
# register_resource(path="/lfc/jquery/jquery.min.js", type=JS)
# register_resource(path="/lfc/jquery/jquery.tools.min.js", type=JS)
# register_resource(path="/lfc/lightbox/js/jquery.lightbox-0.5.js", type=JS)
# register_resource(path="/lfc_theme/js/lfctheme.js", type=JS)



# django imports
from django.conf.urls.defaults import *

urlpatterns = patterns('resources.views',
   url(r'^create-css$', "create_css", name='create_css'),
   url(r'^create-javascript$', "create_javascript", name='create_javascript'),
   url(r'^create-resources$', "create_resources", name='create_resources'),
)


from setuptools import setup, find_packages
import os

version = '1.0a1'

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.txt')).read()

setup(name='django-resources',
      version=version,
      description="CSS and Javascript management for Django.",
      long_description=README,      
      # Get more strings from http://www.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'Framework :: Django',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
        ],
      keywords='django css javascript',
      author='Kai Diefenbach',
      author_email='kai.diefenbach@iqpp.de',
      url='http://www.iqpp.de',
      license='BSD',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=[],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'jsmin',
          'cssutils',
      ],
      )
